@@ -380,9 +380,12 @@ func (s *snapshot) applyCriticalErrorToFiles(ctx context.Context, msg string, fi
 		switch s.view.FileKind(fh) {
 		case source.Go:
 			if pgf, err := s.ParseGo(ctx, fh, source.ParseHeader); err == nil {
-				pkgDecl := span.NewRange(pgf.Tok, pgf.File.Package, pgf.File.Name.End())
-				if spn, err := pkgDecl.Span(); err == nil {
-					rng, _ = pgf.Mapper.Range(spn)
+				// Check that we have a valid `package foo` range to use for positioning the error.
+				if pgf.File.Package.IsValid() && pgf.File.Name != nil && pgf.File.Name.End().IsValid() {
+					pkgDecl := span.NewRange(pgf.Tok, pgf.File.Package, pgf.File.Name.End())
+					if spn, err := pkgDecl.Span(); err == nil {
+						rng, _ = pgf.Mapper.Range(spn)
+					}
 				}
 			}
 		case source.Mod:
